@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2013 Red Hat.
+ * Copyright (c) 2012-2014,2017-2021 Red Hat.
  * Copyright (c) 1995-2002 Silicon Graphics, Inc.  All Rights Reserved.
  * 
  * This program is free software; you can redistribute it and/or modify it
@@ -29,12 +29,10 @@ CheckError(AgentInfo *ap, int sts)
     int		retSts;
 
     if (sts == PM_ERR_PMDANOTREADY) {
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_APPL0)
-	    __pmNotifyErr(LOG_INFO, "%s agent (%s) sent NOT READY\n",
+	if (pmDebugOptions.appl0)
+	    pmNotifyErr(LOG_INFO, "%s agent (%s) sent NOT READY\n",
 			 ap->pmDomainLabel,
 			 ap->status.notReady ? "not ready" : "ready");
-#endif
 	if (ap->status.notReady == 0) {
 	    ap->status.notReady = 1;
 	    retSts = PM_ERR_AGAIN;
@@ -43,12 +41,10 @@ CheckError(AgentInfo *ap, int sts)
 	    retSts = PM_ERR_IPC;
     }
     else if (sts == PM_ERR_PMDAREADY) {
-#ifdef PCP_DEBUG
-	if (pmDebug & DBG_TRACE_APPL0)
-	    __pmNotifyErr(LOG_INFO, "%s agent (%s) sent unexpected READY\n",
+	if (pmDebugOptions.appl0)
+	    pmNotifyErr(LOG_INFO, "%s agent (%s) sent unexpected READY\n",
 			 ap->pmDomainLabel,
 			 ap->status.notReady ? "not ready" : "ready");
-#endif
 	retSts = PM_ERR_IPC;
     }
     else
@@ -69,10 +65,12 @@ DoText(ClientInfo *cp, __pmPDU* pb)
     if ((sts = __pmDecodeTextReq(pb, &ident, &type)) < 0)
 	return sts;
 
-    if ((ap = FindDomainAgent(((__pmID_int *)&ident)->domain)) == NULL)
+    if ((ap = pmcd_agent(((__pmID_int *)&ident)->domain)) == NULL)
 	return PM_ERR_PMID;
-    else if (!ap->status.connected)
+    if (!ap->status.connected)
 	return PM_ERR_NOAGENT;
+    if (ap->status.fenced)
+	return PM_ERR_PMDAFENCED;
 
     if (ap->ipcType == AGENT_DSO) {
 	if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
@@ -87,7 +85,7 @@ DoText(ClientInfo *cp, __pmPDU* pb)
 	sts = __pmSendTextReq(ap->inFd, cp - client, ident, type);
 	if (sts >= 0) {
 	    int		pinpdu;
-	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
 	    if (sts > 0)
 		pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 	    if (sts == PDU_TEXT)
@@ -133,41 +131,28 @@ DoText(ClientInfo *cp, __pmPDU* pb)
 int
 DoProfile(ClientInfo *cp, __pmPDU *pb)
 {
-    __pmProfile	*newProf;
+    __pmHashCtl	*hcp;
+    pmProfile	*newProf;
     int		ctxnum, sts, i;
 
     sts = __pmDecodeProfile(pb, &ctxnum, &newProf);
     if (sts >= 0) {
-	/* Allocate more profile pointers if required */
-	if (ctxnum >= cp->szProfile) {
-	    __pmProfile	**newProfPtrs;
-	    int		need, oldSize = cp->szProfile;
-
-	    if (ctxnum - cp->szProfile < 4)
-		cp->szProfile += 4;
-	    else
-		cp->szProfile = ctxnum + 1;
-	    need = cp->szProfile * (int)sizeof(__pmProfile *);
-	    if ((newProfPtrs = (__pmProfile **)malloc(need)) == NULL) {
-		cp->szProfile = oldSize;
-		__pmNoMem("DoProfile.newProfPtrs", need, PM_RECOV_ERR);
-		__pmFreeProfile(newProf);
-		return -oserror();
+	__pmHashNode	*hp;
+	hcp = &cp->profile;
+	if ((hp = __pmHashSearch(ctxnum, hcp)) != NULL) {
+	    /* seen this context slot before for this client */
+	    pmProfile	*profile = (pmProfile *)hp->data;
+	    if (profile != NULL)
+		__pmFreeProfile(profile);
+	    hp->data = (void *)newProf;
+	}
+	else {
+	    /* first time for this context slot for this client */
+	    if ((sts = __pmHashAdd(ctxnum, newProf, hcp)) > 0) {
+		/* __pmHashAdd returns 1 for success, but we want zero. */
+		sts = 0;
 	    }
-
-	    /* Copy any old pointers and zero the newly allocated ones */
-	    if ((need = oldSize * (int)sizeof(__pmProfile *))) {
-		memcpy(newProfPtrs, cp->profile, need);
-		free(cp->profile);	/* But not the __pmProfile ptrs! */
-	    }
-	    need = (cp->szProfile - oldSize) * (int)sizeof(__pmProfile *);
-	    memset(&newProfPtrs[oldSize], 0, need);
-	    cp->profile = newProfPtrs;
 	}
-	else				/* cp->profile is big enough */
-	    if (cp->profile[ctxnum] != NULL)
-		__pmFreeProfile(cp->profile[ctxnum]);
-	cp->profile[ctxnum] = newProf;
 
 	/* "Invalidate" any references to the client context's profile in the
 	 * agents to which the old profile was last sent
@@ -194,10 +179,12 @@ DoDesc(ClientInfo *cp, __pmPDU *pb)
     if ((sts = __pmDecodeDescReq(pb, &pmid)) < 0)
 	return sts;
 
-    if ((ap = FindDomainAgent(((__pmID_int *)&pmid)->domain)) == NULL)
+    if ((ap = pmcd_agent(((__pmID_int *)&pmid)->domain)) == NULL)
 	return PM_ERR_PMID;
-    else if (!ap->status.connected)
+    if (!ap->status.connected)
 	return PM_ERR_NOAGENT;
+    if (ap->status.fenced)
+	return PM_ERR_PMDAFENCED;
 
     if (ap->ipcType == AGENT_DSO) {
 	if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
@@ -212,7 +199,7 @@ DoDesc(ClientInfo *cp, __pmPDU *pb)
 	sts = __pmSendDescReq(ap->inFd, cp - client, pmid);
 	if (sts >= 0) {
 	    int		pinpdu;
-	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
 	    if (sts > 0)
 		pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 	    if (sts == PDU_DESC)
@@ -257,39 +244,31 @@ DoDesc(ClientInfo *cp, __pmPDU *pb)
 }
 
 int
-DoInstance(ClientInfo *cp, __pmPDU* pb)
+DoInstance(ClientInfo *cp, __pmPDU *pb)
 {
     int			sts, s;
-    __pmTimeval		when;
     pmInDom		indom;
     int			inst;
     char		*name;
-    __pmInResult	*inresult = NULL;
+    pmInResult	*inresult = NULL;
     AgentInfo		*ap;
     int			fdfail = -1;
 
-    sts = __pmDecodeInstanceReq(pb, &when, &indom, &inst, &name);
+    sts = __pmDecodeInstanceReq(pb, &indom, &inst, &name);
     if (sts < 0)
 	return sts;
-    if (when.tv_sec != 0 || when.tv_usec != 0) {
-	/*
-	 * we have no idea how to do anything but current, yet!
-	 *
-	 * TODO EXCEPTION PCP 2.0 ...
-	 * this may be left over from the pmvcr days, and can be tossed?
-	 * ... leaving it here is benign
-	 */
-	if (name != NULL) free(name);
-	return PM_ERR_NYI;
-    }
-    if ((ap = FindDomainAgent(((__pmInDom_int *)&indom)->domain)) == NULL) {
+    if ((ap = pmcd_agent(((__pmInDom_int *)&indom)->domain)) == NULL) {
 	if (name != NULL) free(name);
 	return PM_ERR_INDOM;
     }
-    else if (!ap->status.connected) {
+    if (!ap->status.connected) {
 	if (name != NULL) free(name);
 	return PM_ERR_NOAGENT;
     }
+    if (ap->status.fenced) {
+	if (name != NULL) free(name);
+	return PM_ERR_PMDAFENCED;
+    }
 
     if (ap->ipcType == AGENT_DSO) {
 	if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
@@ -304,10 +283,10 @@ DoInstance(ClientInfo *cp, __pmPDU* pb)
 	    return PM_ERR_AGAIN;
 	}
 	pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_INSTANCE_REQ, (int)indom);
-	sts = __pmSendInstanceReq(ap->inFd, cp - client, &when, indom, inst, name);
+	sts = __pmSendInstanceReq(ap->inFd, cp - client, indom, inst, name);
 	if (sts >= 0) {
 	    int		pinpdu;
-	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
 	    if (sts > 0)
 		pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 	    if (sts == PDU_INSTANCE)
@@ -354,6 +333,246 @@ DoInstance(ClientInfo *cp, __pmPDU* pb)
     return sts;
 }
 
+static int
+GetChangedContextLabels(pmLabelSet **sets, int *changed)
+{
+    int		sts;
+
+    *changed = 0;
+    if ((sts = __pmGetContextLabels(sets)) == 1) {
+	if (!pmcd_labels) {
+	    *changed = 1;
+	} else {
+	    if (strcmp(pmcd_labels, sets[0]->json) != 0)
+		*changed = 1;
+	    free(pmcd_labels);
+	}
+	pmcd_labels = strndup(sets[0]->json, sets[0]->jsonlen);
+    }
+    else if (pmcd_labels) {
+	free(pmcd_labels);
+	pmcd_labels = NULL;
+	*changed = 1;
+    }
+    return sts;
+}
+
+void
+CheckLabelChange(void)
+{
+    pmLabelSet	*sets = NULL;
+    int		nsets;
+
+    if ((nsets = GetChangedContextLabels(&sets, &labelChanged)) > 0)
+	pmFreeLabelSets(sets, nsets);
+}
+
+static int
+GetContextLabels(ClientInfo *cp, pmLabelSet **sets)
+{
+    __pmHashNode	*node;
+    const char		*userid;
+    const char		*groupid;
+    const char		*container;
+    static char		host[MAXHOSTNAMELEN];
+    static char		domain[MAXDOMAINNAMELEN];
+    static char		machineid[MAXMACHINEIDLEN];
+    static const char	*host_label, *domain_label, *machineid_label;
+    char		buf[PM_MAXLABELJSONLEN];
+    int			sts, flags;
+
+    if ((sts = GetChangedContextLabels(sets, &labelChanged)) >= 0) {
+	host_label = __pmGetLabelConfigHostName(host, sizeof(host));
+	domain_label = __pmGetLabelConfigDomainName(domain, sizeof(domain));
+	machineid_label = __pmGetLabelConfigMachineID(machineid, sizeof(machineid));
+	userid = ((node = __pmHashSearch(PCP_ATTR_USERID, &cp->attrs)) ?
+			(const char *)node->data : NULL);
+	groupid = ((node = __pmHashSearch(PCP_ATTR_GROUPID, &cp->attrs)) ?
+			(const char *)node->data : NULL);
+	container = ((node = __pmHashSearch(PCP_ATTR_CONTAINER, &cp->attrs)) ?
+			(const char *)node->data : NULL);
+
+	sts = pmsprintf(buf, sizeof(buf), "{\"%s\":\"%s\"",
+			host_label, host);
+	if (container)
+	    sts += pmsprintf(buf+sts, sizeof(buf)-sts, ",\"container\":\"%s\"",
+			    container);
+	sts += pmsprintf(buf+sts, sizeof(buf)-sts, ",\"%s\":\"%s\"",
+			 domain_label, domain);
+	sts += pmsprintf(buf+sts, sizeof(buf)-sts, ",\"%s\":\"%s\"}",
+			 machineid_label, machineid);
+
+	flags = PM_LABEL_CONTEXT;
+	if ((sts = __pmAddLabels(sets, buf, flags)) <= 0)
+	    return sts;
+
+	flags |= PM_LABEL_OPTIONAL;
+	if (userid && groupid) {
+	    pmsprintf(buf, sizeof(buf), "{\"groupid\":%s,\"userid\":%s}",
+					groupid, userid);
+	    if ((sts = __pmAddLabels(sets, buf, flags)) <= 0)
+		return sts;
+	}
+	return 1;
+    }
+    return sts;
+}
+
+int
+DoLabel(ClientInfo *cp, __pmPDU *pb)
+{
+    __pmHashCtl		*hcp;
+    __pmHashNode	*hp;
+    pmProfile		*profile, defprofile = {PM_PROFILE_INCLUDE, 0, NULL};
+    int			ctxnum;
+    int			sts, s;
+    int			ident, type, nsets = 0;
+    pmLabelSet		*sets = NULL;
+    AgentInfo		*ap = NULL;
+    int			fdfail = -1;
+
+    sts = __pmDecodeLabelReq(pb, &ident, &type);
+    if (sts < 0)
+	return sts;
+
+    switch (type) {
+	case PM_LABEL_CONTEXT:
+	    nsets = sts = GetContextLabels(cp, &sets);
+	    goto response;
+	case PM_LABEL_DOMAIN:
+	    if (!(ap = pmcd_agent(ident)))
+		return PM_ERR_NOAGENT;
+	    break;
+	case PM_LABEL_INDOM:
+	case PM_LABEL_INSTANCES:
+	    if (!(ap = pmcd_agent(((__pmInDom_int *)&ident)->domain)))
+		return PM_ERR_INDOM;
+	    break;
+	case PM_LABEL_CLUSTER:
+	case PM_LABEL_ITEM:
+	    if (!(ap = pmcd_agent(((__pmID_int *)&ident)->domain)))
+		return PM_ERR_PMID;
+	    break;
+	default:
+	    return PM_ERR_TYPE;
+    }
+
+    if (!ap->status.connected)
+	return PM_ERR_NOAGENT;
+    if (ap->status.fenced)
+	return PM_ERR_PMDAFENCED;
+
+    if (ap->ipcType == AGENT_DSO) {
+	if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
+	    ap->ipc.dso.dispatch.version.seven.ext->e_context = cp - client;
+	if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_7) {
+	    sts = ap->ipc.dso.dispatch.version.seven.label(ident, type, &sets,
+					ap->ipc.dso.dispatch.version.any.ext);
+	} else if (type & PM_LABEL_DOMAIN) {
+	    sts = __pmGetDomainLabels(ap->pmDomainId, ap->pmDomainLabel, &sets);
+	    nsets = sts;
+	    goto response;
+	} else {
+	    sts = 0;
+	}
+	if (sts >= 0)
+	    nsets = sts;
+    }
+    else {
+	if (ap->status.notReady)
+	    return PM_ERR_AGAIN;
+
+	/* status.madeDsoResult is only used for DSO agents so don't waste time by
+	 * checking that the agent is a DSO first.
+	 */
+	ap->status.madeDsoResult = 0;
+	ctxnum = cp - client;
+
+	if (ap->profClient != cp || ctxnum != ap->profIndex) {
+	    hcp = &cp->profile;
+	    hp = __pmHashSearch(ctxnum, hcp);
+	    if (hp != NULL)
+		profile = (pmProfile *)hp->data;
+	    else
+		profile = &defprofile;
+	    if (ap->ipcType == AGENT_DSO) {
+		if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
+		    ap->ipc.dso.dispatch.version.four.ext->e_context = cp - client;
+		sts = ap->ipc.dso.dispatch.version.any.profile(profile,
+					 ap->ipc.dso.dispatch.version.any.ext);
+	    }
+	    else {
+		if (ap->status.notReady == 0) {
+		    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PROFILE, ctxnum);
+		    if (profile && (sts = __pmSendProfile(ap->inFd, cp - client,
+					       ctxnum, profile)) < 0) {
+			pmcd_trace(TR_XMIT_ERR, ap->inFd, PDU_PROFILE, sts);
+		    }
+		} else {
+		    sts = PM_ERR_AGAIN;
+		}
+	    }
+	    if (sts >= 0) {
+		ap->profClient = cp;
+		ap->profIndex = ctxnum;
+	    }
+	}
+
+	pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_LABEL_REQ, ident);
+	sts = __pmSendLabelReq(ap->inFd, cp - client, ident, type);
+	if (sts >= 0) {
+	    int		pinpdu;
+	    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
+	    if (sts > 0)
+		pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
+	    if (sts == PDU_LABEL)
+		sts = __pmDecodeLabel(pb, &ident, &type, &sets, &nsets);
+	    else if (sts == PDU_ERROR) {
+		nsets = 0;
+		sets = NULL;
+		s = __pmDecodeError(pb, &sts);
+		if (s < 0)
+		    sts = s;
+		else
+		    sts = CheckError(ap, sts);
+		pmcd_trace(TR_RECV_ERR, ap->outFd, PDU_LABEL, sts);
+	    }
+	    else {
+		pmcd_trace(TR_WRONG_PDU, ap->outFd, PDU_LABEL, sts);
+		sts = PM_ERR_IPC;	/* Wrong PDU type */
+		fdfail = ap->outFd;
+	    }
+	    if (pinpdu > 0)
+		__pmUnpinPDUBuf(pb);
+	}
+	else {
+	    pmcd_trace(TR_XMIT_ERR, ap->inFd, PDU_LABEL_REQ, sts);
+	    fdfail = ap->inFd;
+	}
+    }
+
+response:
+    if (sts >= 0) {
+	pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_LABEL, (int)ident);
+	if (nsets > 1 && !(type & PM_LABEL_INSTANCES))
+	    nsets = 1;
+	sts = __pmSendLabel(cp->fd, FROM_ANON, ident, type, sets, nsets);
+	if (sts < 0) {
+	    pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_LABEL, sts);
+	    CleanupClient(cp, sts);
+	}
+	pmFreeLabelSets(sets, nsets);
+    }
+    else {
+	if (ap && ap->ipcType != AGENT_DSO &&
+	    (sts == PM_ERR_IPC || sts == PM_ERR_TIMEOUT || sts == -EPIPE) &&
+	    fdfail != -1)
+	    CleanupAgent(ap, AT_COMM, fdfail);
+    }
+
+    return sts;
+}
+
 /*
  * This handler is for remote versions of pmNameAll or pmNameID.
  * Note: only one pmid for the list should be sent.
@@ -377,7 +596,7 @@ DoPMNSIDs(ClientInfo *cp, __pmPDU *pb)
 	 * failure may be a real failure, or could be a metric within a
 	 * dynamic sutree of the PMNS
 	 */
-	if ((ap = FindDomainAgent(((__pmID_int *)&idlist[0])->domain)) == NULL) {
+	if ((ap = pmcd_agent(((__pmID_int *)&idlist[0])->domain)) == NULL) {
 	    sts = PM_ERR_NOAGENT;
 	    goto fail;
 	}
@@ -385,6 +604,10 @@ DoPMNSIDs(ClientInfo *cp, __pmPDU *pb)
 	    sts = PM_ERR_NOAGENT;
 	    goto fail;
 	}
+	if (ap->status.fenced) {
+	    sts = PM_ERR_PMDAFENCED;
+	    goto fail;
+	}
 	if (ap->ipcType == AGENT_DSO) {
 	    if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
 		ap->ipc.dso.dispatch.version.four.ext->e_context = cp - client;
@@ -405,7 +628,7 @@ DoPMNSIDs(ClientInfo *cp, __pmPDU *pb)
 	    sts = __pmSendIDList(ap->inFd, cp - client, 1, &idlist[0], 0);
 	    if (sts >= 0) {
 		int		pinpdu;
-		pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+		pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
 		if (sts > 0)
 		    pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 		if (sts == PDU_PMNS_NAMES) {
@@ -435,7 +658,7 @@ DoPMNSIDs(ClientInfo *cp, __pmPDU *pb)
     numnames = sts;
 
     pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, numnames);
-    if ((sts = __pmSendNameList(cp->fd, FROM_ANON, numnames, namelist, NULL)) < 0){
+    if ((sts = __pmSendNameList(cp->fd, FROM_ANON, numnames, (const char **)namelist, NULL)) < 0){
 	pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_PMNS_NAMES, sts);
 	CleanupClient(cp, sts);
     	goto fail;
@@ -457,8 +680,11 @@ fail:
 int
 DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 {
-    int		sts = 0;
+    int		sts;
     int		numids = 0;
+    int		numok;
+    int		lsts;
+    int		domain;
     pmID	*idlist = NULL;
     char	**namelist = NULL;
     int		i;
@@ -472,26 +698,33 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 	goto done;
     }
 
-    sts = pmLookupName(numids, namelist, idlist);
+    sts = pmLookupName(numids, (const char **)namelist, idlist);
+    /*
+     * even if this fails, or looks up fewer than numids, we have to
+     * check each PMID looking for dynamic metrics and process them
+     * separately with the help of the PMDA, if possible
+     */
     for (i = 0; i < numids; i++) {
-	if (idlist[i] == PM_ID_NULL) continue;
-	if (pmid_domain(idlist[i]) == DYNAMIC_PMID && pmid_item(idlist[i]) == 0) {
-	    int		lsts;
-	    int		domain = pmid_cluster(idlist[i]);
-	    /*
-	     * don't return <domain>.*.* ... all return paths from here
-	     * must either set a valid PMID in idlist[i] or indicate
-	     * the first error in the return from pmLookupName
-	     */
-	    idlist[i] = PM_ID_NULL;	/* default case if cannot translate */
-	    if ((ap = FindDomainAgent(domain)) == NULL) {
-		if (sts > 0) sts = PM_ERR_NOAGENT;
-		continue;
-	    }
-	    if (!ap->status.connected) {
-		if (sts > 0) sts = PM_ERR_NOAGENT;
-		continue;
-	    }
+	if (idlist[i] == PM_ID_NULL || !IS_DYNAMIC_ROOT(idlist[i]))
+	    continue;
+	lsts = 0;
+	domain = pmID_cluster(idlist[i]);
+	/*
+	 * don't return <domain>.*.* ... all return paths from here
+	 * must either set a valid PMID in idlist[i] or indicate
+	 * an error via lsts
+	 */
+	idlist[i] = PM_ID_NULL;	/* default case if cannot translate */
+	if ((ap = pmcd_agent(domain)) == NULL) {
+	    lsts = PM_ERR_NOAGENT;
+	}
+	else if (!ap->status.connected) {
+	    lsts = PM_ERR_NOAGENT;
+	}
+	else if (ap->status.fenced) {
+	    lsts = PM_ERR_PMDAFENCED;
+	}
+	else {
 	    if (ap->ipcType == AGENT_DSO) {
 		if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
 		    ap->ipc.dso.dispatch.version.four.ext->e_context = cp - client;
@@ -501,7 +734,7 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 		}
 		else {
 		    /* Not PMDA_INTERFACE_4 or later */
-		    lsts = PM_ERR_NAME;
+		    lsts = PM_ERR_NONLEAF;
 		}
 	    }
 	    else {
@@ -511,10 +744,10 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 		    lsts = PM_ERR_AGAIN;
 		else {
 		    pmcd_trace(TR_XMIT_PDU, ap->inFd, PDU_PMNS_NAMES, 1);
-		    lsts = __pmSendNameList(ap->inFd, cp - client, 1, &namelist[i], NULL);
+		    lsts = __pmSendNameList(ap->inFd, cp - client, 1, (const char **)&namelist[i], NULL);
 		    if (lsts >= 0) {
 			int		pinpdu;
-			pinpdu = lsts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+			pinpdu = lsts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
 			if (lsts > 0)
 			    pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 			if (lsts == PDU_PMNS_IDS) {
@@ -547,28 +780,33 @@ DoPMNSNames(ClientInfo *cp, __pmPDU *pb)
 		    fdfail != -1)
 		    CleanupAgent(ap, AT_COMM, fdfail);
 	    }
-	    /*
-	     * only set error status to the current error status
-	     * if this is the first error for this set of metrics
-	     */
-	    if (lsts < 0 && sts > 0) sts = lsts;
+	}
+	/*
+	 * only set error status to the current error status
+	 * if this is the first error for this set of metrics,
+	 * and if either it is a fatal error, or numids is 1
+	 */
+	if (lsts < 0 && sts > 0) {
+	    if ((lsts != PM_ERR_NAME && lsts != PM_ERR_NOAGENT &&
+		 lsts != PM_ERR_NONLEAF) || numids == 1)
+	    sts = lsts;
 	}
     }
 
-    if (sts < 0) {
-        /* If get an error which should be passed back along
-         * with valid data to the client
-         * then do NOT fail -> return status with the id-list.
-         */
-    	if (sts != PM_ERR_NAME && sts != PM_ERR_NONLEAF && sts != PM_ERR_NOAGENT)
-	  goto done;
+    if (sts < 0)
+	/* fatal error or explicit error in the numids == 1 case */
+	goto done;
+
+    numok = numids;
+    for (i = 0; i < numids; i++) {
+	if (idlist[i] == PM_ID_NULL)
+	    numok--;
     }
 
-    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_IDS, numids);
-    if ((sts = __pmSendIDList(cp->fd, FROM_ANON, numids, idlist, sts)) < 0) {
+    pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_IDS, numok);
+    if ((sts = __pmSendIDList(cp->fd, FROM_ANON, numids, idlist, numok)) < 0) {
 	pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_PMNS_IDS, sts);
 	CleanupClient(cp, sts);
-    	goto done;
     }
 
 done:
@@ -597,11 +835,11 @@ DoPMNSChild(ClientInfo *cp, __pmPDU *pb)
 	goto done;
 
     namelist[0] = name;
-    sts = pmLookupName(1, namelist, idlist);
-    if (sts == 1 && pmid_domain(idlist[0]) == DYNAMIC_PMID && pmid_item(idlist[0]) == 0) {
-	int		domain = pmid_cluster(idlist[0]);
+    sts = pmLookupName(1, (const char **)namelist, idlist);
+    if (sts == 1 && IS_DYNAMIC_ROOT(idlist[0])) {
+	int		domain = pmID_cluster(idlist[0]);
 	AgentInfo	*ap = NULL;
-	if ((ap = FindDomainAgent(domain)) == NULL) {
+	if ((ap = pmcd_agent(domain)) == NULL) {
 	    sts = PM_ERR_NOAGENT;
 	    goto done;
 	}
@@ -609,6 +847,10 @@ DoPMNSChild(ClientInfo *cp, __pmPDU *pb)
 	    sts = PM_ERR_NOAGENT;
 	    goto done;
 	}
+	if (ap->status.fenced) {
+	    sts = PM_ERR_PMDAFENCED;
+	    goto done;
+	}
 	if (ap->ipcType == AGENT_DSO) {
 	    if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
 		ap->ipc.dso.dispatch.version.four.ext->e_context = cp - client;
@@ -638,7 +880,7 @@ DoPMNSChild(ClientInfo *cp, __pmPDU *pb)
 		sts = __pmSendChildReq(ap->inFd, cp - client, name, subtype);
 		if (sts >= 0) {
 		    int		pinpdu;
-		    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+		    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
 		    if (sts > 0)
 			pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 		    if (sts == PDU_PMNS_NAMES) {
@@ -689,7 +931,7 @@ DoPMNSChild(ClientInfo *cp, __pmPDU *pb)
 
     numnames = sts;
     pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, numnames);
-    if ((sts = __pmSendNameList(cp->fd, FROM_ANON, numnames, offspring, statuslist)) < 0) {
+    if ((sts = __pmSendNameList(cp->fd, FROM_ANON, numnames, (const char **)offspring, statuslist)) < 0) {
 	pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_PMNS_NAMES, sts);
 	CleanupClient(cp, sts);
     }
@@ -712,16 +954,16 @@ static int travNL_i;      /* array index */
 static void
 AddLengths(const char *name)
 {
-  travNL_strlen += strlen(name) + 1;
-  travNL_num++;
+    travNL_strlen += strlen(name) + 1;
+    travNL_num++;
 }
 
 static void
 BuildNameList(const char *name)
 {
-  travNL[travNL_i++] = travNL_ptr;
-  strcpy(travNL_ptr, name);
-  travNL_ptr += strlen(name) + 1;
+    travNL[travNL_i++] = travNL_ptr;
+    strcpy(travNL_ptr, name);
+    travNL_ptr += strlen(name) + 1;
 }
 
 /*
@@ -772,24 +1014,25 @@ traverse_dynamic(ClientInfo *cp, char *start, int *num_names, char ***names)
     for (i = *num_names-1; i >= 0; i--) {
 	offspring = NULL;
 	namelist[0] = (*names)[i];
-	sts = pmLookupName(1, namelist, idlist);
+	sts = pmLookupName(1, (const char **)namelist, idlist);
 	if (sts < 1)
 	    continue;
-	if (pmid_domain(idlist[0]) == DYNAMIC_PMID && pmid_item(idlist[0]) == 0) {
-	    int		domain = pmid_cluster(idlist[0]);
+	if (IS_DYNAMIC_ROOT(idlist[0])) {
+	    int		domain = pmID_cluster(idlist[0]);
 	    AgentInfo	*ap;
-	    if ((ap = FindDomainAgent(domain)) == NULL)
+	    if ((ap = pmcd_agent(domain)) == NULL)
 		continue;
 	    if (!ap->status.connected)
 		continue;
+	    if (ap->status.fenced)
+		continue;
 	    if (ap->ipcType == AGENT_DSO) {
 		if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_5)
 		    ap->ipc.dso.dispatch.version.four.ext->e_context = cp - client;
 		if (ap->ipc.dso.dispatch.comm.pmda_interface >= PMDA_INTERFACE_4) {
 		    sts = ap->ipc.dso.dispatch.version.four.children(namelist[0], 1, &offspring, &statuslist,
 				      ap->ipc.dso.dispatch.version.four.ext);
-#ifdef PCP_DEBUG
-		    if (pmDebug & DBG_TRACE_PMNS) {
+		    if (pmDebugOptions.pmns) {
 			fprintf(stderr, "traverse_dynamic: DSO PMDA: expand dynamic PMNS entry %s (%s) -> ", namelist[0], pmIDStr(idlist[0]));
 			if (sts < 0)
 			    fprintf(stderr, "%s\n", pmErrStr(sts));
@@ -801,7 +1044,6 @@ traverse_dynamic(ClientInfo *cp, char *start, int *num_names, char ***names)
 			    }
 			}
 		    }
-#endif
 		    if (sts < 0)
 			continue;
 		    if (statuslist) free(statuslist);
@@ -822,14 +1064,13 @@ traverse_dynamic(ClientInfo *cp, char *start, int *num_names, char ***names)
 		    int		numnames;
 		    __pmPDU	*pb;
 		    int		pinpdu;
-		    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, _pmcd_timeout, &pb);
+		    pinpdu = sts = __pmGetPDU(ap->outFd, ANY_SIZE, pmcd_timeout, &pb);
 		    if (sts > 0)
 			pmcd_trace(TR_RECV_PDU, ap->outFd, sts, (int)((__psint_t)pb & 0xffffffff));
 		    if (sts == PDU_PMNS_NAMES) {
 			sts = __pmDecodeNameList(pb, &numnames,
 						       &offspring, &statuslist);
-#ifdef PCP_DEBUG
-			if (pmDebug & DBG_TRACE_PMNS) {
+			if (pmDebugOptions.pmns) {
 			    fprintf(stderr, "traverse_dynamic: daemon PMDA: expand dynamic PMNS entry %s (%s) -> ", namelist[0], pmIDStr(idlist[0]));
 			    if (sts < 0)
 				fprintf(stderr, "%s\n", pmErrStr(sts));
@@ -841,7 +1082,6 @@ traverse_dynamic(ClientInfo *cp, char *start, int *num_names, char ***names)
 				}
 			    }
 			}
-#endif
 			if (statuslist) {
 			    free(statuslist);
 			    statuslist = NULL;
@@ -956,11 +1196,9 @@ DoPMNSTraverse(ClientInfo *cp, __pmPDU *pb)
     travNL_num = 0;
     if ((sts = pmTraversePMNS(name, AddLengths)) < 0)
     	goto check;
-#ifdef PCP_DEBUG
-    if (pmDebug & DBG_TRACE_PMNS) {
+    if (pmDebugOptions.pmns) {
 	fprintf(stderr, "DoPMNSTraverse: %d names below %s after pmTraversePMNS\n", travNL_num, name);
     }
-#endif
 
     /* for each ptr, string bytes, and string terminators */
     travNL_need = travNL_num * (int)sizeof(char*) + travNL_strlen;
@@ -986,7 +1224,7 @@ check:
 	goto done;
 
     pmcd_trace(TR_XMIT_PDU, cp->fd, PDU_PMNS_NAMES, travNL_num);
-    if ((sts = __pmSendNameList(cp->fd, FROM_ANON, travNL_num, travNL, NULL)) < 0) {
+    if ((sts = __pmSendNameList(cp->fd, FROM_ANON, travNL_num, (const char **)travNL, NULL)) < 0) {
 	pmcd_trace(TR_XMIT_ERR, cp->fd, PDU_PMNS_NAMES, sts);
 	CleanupClient(cp, sts);
 	goto done;
@@ -1009,7 +1247,7 @@ GetAttribute(ClientInfo *cp, int code)
     int		sts, pinpdu;
 
     /* Expecting an attribute (code) PDU from the client */
-    pinpdu = sts = __pmGetPDU(cp->fd, LIMIT_SIZE, _pmcd_timeout, &pb);
+    pinpdu = sts = __pmGetPDU(cp->fd, LIMIT_SIZE, pmcd_timeout, &pb);
     if (sts > 0)
 	pmcd_trace(TR_RECV_PDU, cp->fd, sts, (int)((__psint_t)pb & 0xffffffff));
     if (sts == PDU_ATTR) {
@@ -1036,7 +1274,7 @@ ConnectionAttributes(ClientInfo *cp, int flags)
     int sts;
 
     if ((sts = __pmSecureServerHandshake(cp->fd, flags, &cp->attrs)) < 0) {
-	if (pmDebug & DBG_TRACE_AUTH)
+	if (pmDebugOptions.auth)
 	    fprintf(stderr, "DoCreds: __pmSecureServerHandshake gave %d: %s\n",
 		    sts, pmErrStr(sts));
 	return sts;
@@ -1044,7 +1282,7 @@ ConnectionAttributes(ClientInfo *cp, int flags)
 
     if ((flags & PDU_FLAG_CONTAINER) &&
 	(sts = GetAttribute(cp, PCP_ATTR_CONTAINER)) < 0) {
-	if (pmDebug & DBG_TRACE_ATTR)
+	if (pmDebugOptions.attr)
 	    fprintf(stderr, "DoCreds: failed GetAttribute container %d: %s\n",
 		    sts, pmErrStr(sts));
 	return sts;
@@ -1053,6 +1291,16 @@ ConnectionAttributes(ClientInfo *cp, int flags)
     return 0;
 }
 
+static int
+CheckCertRequired(ClientInfo *cp)
+{
+    if(__pmServerHasFeature(PM_SERVER_FEATURE_CERT_REQD))
+	if ( !__pmSockAddrIsLoopBack(cp->addr) && !__pmSockAddrIsUnix(cp->addr) )
+	    return 1;
+
+    return 0;
+}
+
 int
 DoCreds(ClientInfo *cp, __pmPDU *pb)
 {
@@ -1070,8 +1318,7 @@ DoCreds(ClientInfo *cp, __pmPDU *pb)
 		vcp = (__pmVersionCred *)&credlist[i];
 		flags = vcp->c_flags;
 		version = vcp->c_version;
-#ifdef PCP_DEBUG
-		if (pmDebug & DBG_TRACE_ATTR) {
+		if (pmDebugOptions.attr) {
 		    static const struct {
 			int	flag;
 			char	*name;
@@ -1083,29 +1330,29 @@ DoCreds(ClientInfo *cp, __pmPDU *pb)
 			{ PDU_FLAG_SECURE_ACK,	"SECURE_ACK" },
 			{ PDU_FLAG_NO_NSS_INIT,	"NO_NSS_INIT" },
 			{ PDU_FLAG_CONTAINER,	"CONTAINER" },
+			{ PDU_FLAG_BAD_LABEL,	"BAD_LABEL" },
+			{ PDU_FLAG_LABELS,	"LABELS" },
+			{ PDU_FLAG_HIGHRES,	"HIGHRES" },
 		    };
-		    int	i;
+		    int	n;
 		    int	first = 1;
 		    fprintf(stderr, "DoCreds: version cred (%u) flags=%x (", vcp->c_version, vcp->c_flags);
-		    for (i = 0; i < sizeof(flag_dbg)/sizeof(flag_dbg[0]); i++) {
-			if (flags & flag_dbg[i].flag) {
+		    for (n = 0; n < sizeof(flag_dbg)/sizeof(flag_dbg[0]); n++) {
+			if (flags & flag_dbg[n].flag) {
 			    if (first)
 				first = 0;
 			    else
 				fputc('|', stderr);
-			    fprintf(stderr, "%s", flag_dbg[i].name);
+			    fprintf(stderr, "%s", flag_dbg[n].name);
 			}
 		    }
 		    fprintf(stderr, ")\n");
 		}
-#endif
 		break;
 
 	    default:
-#ifdef PCP_DEBUG
-		if (pmDebug & DBG_TRACE_AUTH)
+		if (pmDebugOptions.auth)
 		    fprintf(stderr, "DoCreds: Error: bogus cred type %d\n", credlist[i].c_type);
-#endif
 		sts = PM_ERR_IPC;
 		break;
 	}
@@ -1115,6 +1362,18 @@ DoCreds(ClientInfo *cp, __pmPDU *pb)
 
     if (sts >= 0 && version)
 	sts = __pmSetVersionIPC(cp->fd, version);
+
+    /*
+     * In normal operation, some of this code is redundant. A 
+     * remote client should error out during initial handshake
+     * if it does not support client certs.
+     *
+     * We still need to check local connections and allow those through
+     * in all cases.
+     */
+    if (CheckCertRequired(cp) && (flags & PDU_FLAG_SECURE) == 0)
+	return PM_ERR_NEEDCLIENTCERT;
+
     if (sts >= 0 && flags) {
 	/*
 	 * new client has arrived; may want encryption, authentication, etc
@@ -1126,7 +1385,7 @@ DoCreds(ClientInfo *cp, __pmPDU *pb)
 	    return sts;
     }
     if ((sts = CheckAccountAccess(cp)) < 0) {	/* host access done already */
-	if (pmDebug & DBG_TRACE_AUTH)
+	if (pmDebugOptions.auth)
 	    fprintf(stderr, "DoCreds: CheckAccountAccess returns %d: %s\n",
 		    sts, pmErrStr(sts));
 	return sts;
@@ -1137,7 +1396,7 @@ DoCreds(ClientInfo *cp, __pmPDU *pb)
      */
     else if (sts > 0 || (flags & PDU_FLAG_CONTAINER)) {
 	sts = AgentsAttributes(cp - client);
-	if (sts < 0 && (pmDebug & (DBG_TRACE_AUTH|DBG_TRACE_ATTR)))
+	if (sts < 0 && (pmDebugOptions.auth || pmDebugOptions.attr))
 	    fprintf(stderr, "DoCreds: AgentsAttributes returns %d: %s\n",
 		    sts, pmErrStr(sts));
     }
